package com.chelvc.framework.rocketmq.fallback;

import java.io.IOException;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import com.chelvc.framework.base.context.Session;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ThreadUtils;
import com.chelvc.framework.rocketmq.context.RocketMQContextHolder;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.rocketmq.client.apis.message.Message;
import org.apache.rocketmq.client.apis.producer.Producer;
import org.apache.rocketmq.client.apis.producer.Transaction;

/**
 * RocketMQ降级内存消息生产者实现
 *
 * @author Woody
 * @date 2024/7/14
 */
@Slf4j
public class RocketMQMemoryProducer extends RocketMQFallbackProducer {
    private final int capacity;
    private volatile Thread thread;
    private volatile BlockingQueue<Pair<Pair<Message, Transaction>, Session>> queue;

    public RocketMQMemoryProducer(@NonNull Producer delegate) {
        this(10000, delegate);
    }

    public RocketMQMemoryProducer(int capacity, @NonNull Producer delegate) {
        super(delegate);
        AssertUtils.check(capacity > 0, () -> "RocketMQ fallback queue capacity must be greater than 0");
        this.capacity = capacity;
    }

    @Override
    protected void fallback(Message message, Transaction transaction) {
        if (log.isDebugEnabled()) {
            String topic = RocketMQContextHolder.topic(message), body = RocketMQContextHolder.body(message);
            log.debug("RocketMQ fallback message: {}, {}", topic, body);
        }

        // 初始化消息队列及重试线程
        if (this.queue == null) {
            synchronized (this) {
                if (this.queue == null) {
                    this.queue = new ArrayBlockingQueue<>(this.capacity);
                    this.thread = ThreadUtils.consume(this.queue, pair -> {
                        Message _message = pair.getLeft().getLeft();
                        Transaction _transaction = pair.getLeft().getRight();
                        SessionContextHolder.setSession(pair.getRight());
                        try {
                            if (_transaction == null) {
                                this.delegate.send(_message);
                            } else {
                                this.delegate.send(_message, _transaction);
                                _transaction.commit();
                            }
                            return true;
                        } catch (Exception e) {
                            log.error("RocketMQ message send failed: {}", RocketMQContextHolder.topic(_message), e);
                        } finally {
                            SessionContextHolder.clearSessionContext();
                        }
                        ThreadUtils.sleep(5000);
                        return false;
                    });
                }
            }
        }

        // 将消息放入内存消息队列
        Session session = SessionContextHolder.getSession(false);
        if (!this.queue.offer(Pair.of(Pair.of(message, transaction), session))) {
            String topic = RocketMQContextHolder.topic(message), body = RocketMQContextHolder.body(message);
            throw new RuntimeException("RocketMQ fallback queue is full: " + topic + ", " + body);
        }
    }

    @Override
    public void close() throws IOException {
        try {
            if (this.thread != null) {
                this.thread.interrupt();
                ThreadUtils.join(this.thread);
            }
        } finally {
            super.close();
        }
    }
}
